local *

{
  :CreateFrame
  :Elements
  :GetRuneCooldown
  :LibStub
  :UIParent
  :UnitFrame
  :idTweaks
  :idTweaksModule
  :idTweaksUtils
  :oUF
} = _G

LibSharedMedia = LibStub "LibSharedMedia-3.0"
shared_media_status_bars = LibSharedMedia\HashTable "statusbar"
shared_media_status_bars_reversed = idTweaksUtils.table.map.reverse shared_media_status_bars
shared_media_fonts = LibSharedMedia\HashTable "font"
shared_media_fonts_reversed = idTweaksUtils.table.map.reverse shared_media_fonts
shared_media_backgrounds = LibSharedMedia\HashTable "background"
shared_media_backgrounds_reversed = idTweaksUtils.table.map.reverse shared_media_backgrounds

TL, TC, TR = "TOPLEFT", "TOP", "TOPRIGHT"
ML, MC, MR = "LEFT", "CENTER", "RIGHT"
BL, BC, BR = "BOTTOMLEFT", "BOTTOM", "BOTTOMRIGHT"

oUF.Tags.Methods.runes = (unit, real_unit) ->
  return unless unit == "player"
  amount = 0
  for i = 1, 6 do
    _, _, ready = GetRuneCooldown i
    amount = amount + 1 if ready == true
  amount

class UnitFrames extends idTweaksModule
  default:
    enabled: true
    frame:
      size:
        width: 250
        height: 40
        padding: 0
        margin: 0
      -- TODO: Implement.
      positioning:
        from: "CENTER"
        target: "UIParent"
        to: "CENTER"
        x: 0
        y: 0
      -- TODO: Implement.
      background_texture: ""
      -- TODO: Implement.
      border_texture: ""
      border_insets:
        top: 0
        left: 0
        right: 0
        bottom: 0
    bars:
      health_bar:
        enabled: true
        height: 0.6
        texture: "Interface\\TargetingFrame\\UI-StatusBar"
        -- TODO: Rename to background_texture
        background: "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
        -- TODO: Implement.
        border_texture: ""
        border_insets:
          top: 0
          left: 0
          right: 0
          bottom: 0
      power_bar:
        enabled: true
        height: 0.2
        texture: "Interface\\TargetingFrame\\UI-StatusBar"
        -- TODO: Rename to background_texture
        background: "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
        -- TODO: Implement.
        border_texture: ""
        border_insets:
          top: 0
          left: 0
          right: 0
          bottom: 0
      cast_bar:
        enabled: true
        height: 0.2
        texture: "Interface\\TargetingFrame\\UI-StatusBar"
        -- TODO: Rename to background_texture
        background: "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
        -- TODO: Implement.
        border_texture: ""
        border_insets:
          top: 0
          left: 0
          right: 0
          bottom: 0
      additional_power_bar:
        enabled: false
        height: 0.2
        texture: "Interface\\TargetingFrame\\UI-StatusBar"
        -- TODO: Rename to background_texture
        background: "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
        -- TODO: Implement.
        border_texture: ""
        border_insets:
          top: 0
          left: 0
          right: 0
          bottom: 0
    buffs:
      buffs:
        rows: 2
        columns: 8
      debuffs:
        rows: 2
        columns: 8
    text:
      name:
        padding: 5
        font: "Fonts\\FRIZQT__.TTF"
        color:
          r: 1.0
          g: 1.0
          b: 1.0
          a: 1.0
        tag: "[name][leader][status][threatcolor][threat]"
      health:
        padding: 5
        font: "Fonts\\FRIZQT__.TTF"
        color:
          r: 1.0
          g: 1.0
          b: 1.0
          a: 1.0
        tag: "[curhp]"
      classpower:
        padding: 5
        font: "Fonts\\FRIZQT__.TTF"
        color:
          r: 1.0
          g: 1.0
          b: 1.0
          a: 1.0
        tag: "[runes][arcanecharges][chi][holypower][soulshards][cpoints]"
      power:
        padding: 5
        font: "Fonts\\FRIZQT__.TTF"
        color:
          r: 1.0
          g: 1.0
          b: 1.0
          a: 1.0
        tag: "[curpp]"

  new: (addon) =>
    -- print "UnitFrames#new"

    super addon, "UnitFrames", "AceConsole-3.0", "AceEvent-3.0"

    @frames = {}
    @frames1 = {}
    @size = {}

  -- TODO: enabled_units
  units: { "pet", "player", "target", "targettarget" }
  set_defaults: =>
    -- print "UnitFrames#set_defaults"

    @defaults =
      profile:
        enabled: true
        layout: "horizontal"
        default: @default
        pet: idTweaksUtils.table.map.merge @default,
          defaults: true
        player: idTweaksUtils.table.map.merge @default,
          defaults: true
        target: idTweaksUtils.table.map.merge @default,
          defaults: true
        targettarget: idTweaksUtils.table.map.merge @default,
          defaults: true
        party: idTweaksUtils.table.map.merge @default,
          defaults: true

  get_unit_options: (unit, unit_name) =>
    -- print "UnitFrames#get_unit_options #{unit} #{unit_name}"

    unit_has_defaults = (unit) ->
      ->
        not not @db[unit].defaults

    @create_group unit, unit_name, unit_name,
      defaults: @create_toggle "#{unit}.defaults", "Defaults", "Use Default Unit options for this unit",
        order: 0

      frame: @create_group "#{unit}.frame", "Frame", "Settings that affect the Frame",
        order: 1
        disabled: unit_has_defaults unit

        size: @create_group "#{unit}.frame.size", "Size", "Control the size of the frame",
          order: 1
          disabled: unit_has_defaults unit

          width: @create_range "#{unit}.frame.size.width", "Width", "Width of the frame",
            order: 1
            min: 100
            max: 1000
            step: 1
            bigStep: 10
            set: (_, value) ->
              @db[unit].frame.size.width = value
              for i, unit in ipairs @units
                @size_position_frame unit, @frames[unit]
                @size_position_elements unit, @frames[unit]

          height: @create_range "#{unit}.frame.size.height", "Height", "Height of the frame",
            order: 2
            min: 5
            max: 100
            step: 1
            bigStep: 1
            set: (_, value) ->
              @db[unit].frame.size.height = value
              for i, unit in ipairs @units
                @size_position_frame unit, @frames[unit]
                @size_position_elements unit, @frames[unit]

          padding: @create_range "#{unit}.frame.size.padding", "Padding", "Padding of the frame",
            order: 3
            min: 0
            max: 100
            step: 1
            bigStep: 1
            set: (_, value) ->
              @db[unit].frame.size.padding = value
              for i, unit in ipairs @units
                @size_position_frame unit, @frames[unit]
                @size_position_elements unit, @frames[unit]

          margin: @create_range "#{unit}.frame.size.margin", "Margin", "Margin of the frame",
            order: 4
            min: 0
            max: 100
            step: 1
            bigStep: 1
            set: (_, value) ->
              @db[unit].frame.size.margin = value
              for i, unit in ipairs @units
                @size_position_frame unit, @frames[unit]
                @size_position_elements unit, @frames[unit]

        positioning: @create_group "#{unit}.frame.positioning", "Positioning", "Control the positioning of the frame",
          order: 2
          disabled: unit_has_defaults unit

      bars: @create_group "#{unit}.bars", "Bars", "Settings that affect the Bars",
        order: 2
        disabled: unit_has_defaults unit

        health_bar: @create_group "#{unit}.bars.health_bar", "Health Bar", "Health Bar",
          order: 1
          disabled: unit_has_defaults unit

          enabled: @create_toggle "#{unit}.bars.health_bar.enabled", "Enabled", "Turns the bar on and off",
            order: 0

          texture: @create_select "#{unit}.bars.health_bar.texture", "Texture", "Texture",
            order: 1
            dialogControl: "LSM30_Statusbar"
            values: LibSharedMedia\HashTable("statusbar")
            set: (_, value) ->
              @db[unit].bars.health_bar.texture = shared_media_status_bars[value]
              @update_all_units!

          background: @create_select "#{unit}.bars.health_bar.background", "Background", "Background",
            order: 2
            dialogControl: "LSM30_Statusbar"
            values: LibSharedMedia\HashTable("statusbar")
            set: (_, value) ->
              @db[unit].bars.health_bar.background = shared_media_status_bars[value]
              @update_all_units!

          height: @create_range "#{unit}.bars.health_bar.height", "Height", "Height of the frame",
            order: 3
            min: 0.1
            max: 2.0
            step: 0.1
            bigStep: 0.5
            set: (_, value) ->
              @db[unit].bars.health_bar.height = value
              @update_all_units!

        power_bar: @create_group "#{unit}.bars.power_bar", "Power Bar", "Power Bar",
          order: 2
          disabled: unit_has_defaults unit

          enabled: @create_toggle "#{unit}.bars.power_bar.enabled", "Enabled", "Turns the bar on and off",
            order: 0

          texture: @create_select "#{unit}.bars.power_bar.texture", "Texture", "Texture",
            order: 1
            dialogControl: "LSM30_Statusbar"
            values: LibSharedMedia\HashTable("statusbar")
            set: (_, value) ->
              @db[unit].bars.power_bar.texture = shared_media_status_bars[value]
              @update_all_units!

          background: @create_select "#{unit}.bars.power_bar.background", "Background", "Background",
            order: 2
            dialogControl: "LSM30_Statusbar"
            values: LibSharedMedia\HashTable("statusbar")
            set: (_, value) ->
              @db[unit].bars.power_bar.background = shared_media_status_bars[value]
              @update_all_units!

          height: @create_range "#{unit}.bars.power_bar.height", "Height", "Height of the frame",
            order: 3
            min: 0.1
            max: 2.0
            step: 0.1
            bigStep: 0.5
            set: (_, value) ->
              @db[unit].bars.power_bar.height = value
              @update_all_units!

        cast_bar: @create_group "#{unit}.bars.cast_bar", "Cast Bar", "Cast Bar",
          order: 2

          enabled: @create_toggle "#{unit}.bars.cast_bar.enabled", "Enabled", "Turns the bar on and off",
            order: 1

          -- texture: @create_select "#{unit}.bars.cast_bar.texture", "Texture", "Texture",
          --   order: 2
          --   dialogControl: "LSM30_Statusbar"
          --   values: LibSharedMedia\HashTable("statusbar")
          --   set: (_, value) ->
          --     @db[unit].bars.cast_bar.texture = shared_media_status_bars[value]
          --     @update_all_units!

          -- background: @create_select "#{unit}.bars.cast_bar.background", "Background", "Background",
          --   order: 3
          --   dialogControl: "LSM30_Statusbar"
          --   values: LibSharedMedia\HashTable("statusbar")
          --   set: (_, value) ->
          --     @db[unit].bars.cast_bar.background = shared_media_status_bars[value]
          --     @update_all_units!

          -- height: @create_range "#{unit}.bars.cast_bar.height", "Height", "Height of the frame",
          --   order: 4
          --   min: 0.1
          --   max: 2.0
          --   step: 0.1
          --   bigStep: 0.5
          --   set: (_, value) ->
          --     @db[unit].bars.cast_bar.height = value
          --     @update_all_units!

        additional_power_bar: @create_group "#{unit}.bars.additional_power_bar", "Additional Power Bar", "Additional Power Bar",
          order: 2
          disabled: unit_has_defaults unit

          enabled: @create_toggle "#{unit}.bars.additional_power_bar.enabled", "Enabled", "Turns the bar on and off",
            order: 0

          texture: @create_select "#{unit}.bars.additional_power_bar.texture", "Texture", "Texture",
            order: 1
            dialogControl: "LSM30_Statusbar"
            values: LibSharedMedia\HashTable("statusbar")
            set: (_, value) ->
              @db[unit].bars.additional_power_bar.texture = shared_media_status_bars[value]
              @update_all_units!

          background: @create_select "#{unit}.bars.additional_power_bar.background", "Background", "Background",
            order: 2
            dialogControl: "LSM30_Statusbar"
            values: LibSharedMedia\HashTable("statusbar")
            set: (_, value) ->
              @db[unit].bars.additional_power_bar.background = shared_media_status_bars[value]
              @update_all_units!

          height: @create_range "#{unit}.bars.additional_power_bar.height", "Height", "Height of the frame",
            order: 3
            min: 0.1
            max: 2.0
            step: 0.1
            bigStep: 0.5
            set: (_, value) ->
              @db[unit].bars.additional_power_bar.height = value
              @update_all_units!

      buffs: @create_group "#{unit}.buffs", "Buffs", "Settings that affect the Buffs",
        order: 3
        disabled: unit_has_defaults unit

        buffs: @create_group "#{unit}.buffs.buffs", "Buffs", "Buffs",
          order: 1
          disabled: unit_has_defaults unit

          rows: @create_range "#{unit}.buffs.buffs.rows", "Rows", "Number of Rows",
            order: 1
            min: 1
            max: 10
            step: 1
            bigStep: 5
            set: (_, value) ->
              @db[unit].buffs.buffs.rows = value
              @update_all_units!

          columns: @create_range "#{unit}.buffs.buffs.columns", "Columns", "Number of Columns",
            order: 1
            min: 1
            max: 10
            step: 1
            bigStep: 5
            set: (_, value) ->
              @db[unit].buffs.buffs.columns = value
              @update_all_units!

        debuffs: @create_group "#{unit}.buffs.debuffs", "Debuffs", "Debuffs",
          order: 2
          disabled: unit_has_defaults unit

          rows: @create_range "#{unit}.buffs.debuffs.rows", "Rows", "Number of Rows",
            order: 1
            min: 1
            max: 10
            step: 1
            bigStep: 5
            set: (_, value) ->
              @db[unit].buffs.debuffs.rows = value
              @update_all_units!

          columns: @create_range "#{unit}.buffs.debuffs.columns", "Columns", "Number of Columns",
            order: 1
            min: 1
            max: 10
            step: 1
            bigStep: 5
            set: (_, value) ->
              @db[unit].buffs.debuffs.columns = value
              @update_all_units!

      text: @create_group "#{unit}.text", "Text", "Text",
        order: 4
        disabled: unit_has_defaults unit

        name: @create_group "#{unit}.text.name", "Name", "Name",
          order: 1

          padding: @create_range "#{unit}.text.name.padding", "Padding", "Padding",
            order: 1
            min: 1
            max: 10
            step: 1
            bigStep: 5
            set: (_, value) ->
              @db[unit].text.name.padding = value
              @update_all_units!

          font: @create_select "#{unit}.text.name.font", "Font", "Font",
            order: 1
            dialogControl: "LSM30_Font"
            values: shared_media_fonts
            set: (_, value) ->
              @db[unit].text.name.font = shared_media_fonts[value]
              @update_all_units!

          -- tag: @create_input "#{unit}.text.name.tag", "Tag", "Tag",
          --   order: 3

        -- health: @create_group "#{unit}.text.health", "Health", "Health",
        --   order: 1

        --   padding: @create_range "#{unit}.text.health.padding", "Padding", "Padding",
        --     order: 1
        --     min: 1
        --     max: 10
        --     step: 1
        --     bigStep: 5
        --     set: (_, value) ->
        --       @db[unit].text.name.padding = value
        --       @update_all_units!

        --   font: @create_input "#{unit}.text.health.font", "Font", "Font",
        --     order: 2

        --   tag: @create_input "#{unit}.text.health.tag", "Tag", "Tag",
        --     order: 3

        -- classpower: @create_group "#{unit}.text.classpower", "Class Power", "Class Power",
        --   order: 1

        --   padding: @create_range "#{unit}.text.classpower.padding", "Padding", "Padding",
        --     order: 1
        --     min: 1
        --     max: 10
        --     step: 1
        --     bigStep: 5
        --     set: (_, value) ->
        --       @db[unit].text.name.padding = value
        --       @update_all_units!

        --   font: @create_input "#{unit}.text.classpower.font", "Font", "Font",
        --     order: 2

        --   tag: @create_input "#{unit}.text.classpower.tag", "Tag", "Tag",
        --     order: 3

        -- power: @create_group "#{unit}.text.power", "Power", "Power",
        --   order: 1

        --   padding: @create_range "#{unit}.text.power.padding", "Padding", "Padding",
        --     order: 1
        --     min: 1
        --     max: 10
        --     step: 1
        --     bigStep: 5
        --     set: (_, value) ->
        --       @db[unit].text.name.padding = value
        --       @update_all_units!

        --   font: @create_input "#{unit}.text.power.font", "Font", "Font",
        --     order: 2

        --   tag: @create_input "#{unit}.text.power.tag", "Tag", "Tag",
        --     order: 3

  set_options: =>
    print "UnitFrames#set_options"

    super!

    @options.args.layout = @create_select "layout", "Layout", "Switch layouts",
      order: 1
      values: {
        horizontal: "Horizontal",
        vertical: "Vertical",
      }
      set: (_, value) ->
        @db.layout = value
        @update_all_units!

    @options.args.default = @get_unit_options "default", "Default"
    @options.args.default.args.defaults = nil

    @options.args.pet = @get_unit_options "pet", "Pet"
    @options.args.player = @get_unit_options "player", "Player"
    @options.args.target = @get_unit_options "target", "Target"
    @options.args.targettarget = @get_unit_options "targettarget", "Targettarget"

  initialize: =>
    -- print "UnitFrames#initialize"

    oUF\RegisterStyle "idTweaks_default", (frame, unit) -> @layout_default unit, frame
    oUF\SetActiveStyle "idTweaks_default"

  enable: =>
    -- print "UnitFrames#enable"

    @frames.player = oUF\Spawn "player"
    @frames.pet = oUF\Spawn "pet"
    @frames.target = oUF\Spawn "target"
    @frames.targettarget = oUF\Spawn "targettarget"

    add_frame = (unit) ->
      db = @get_db_for_unit unit

      @frames1[unit] = UnitFrame unit

      for k, v in pairs db.bars
        @frames1[unit]\add_bar Elements.bars[v]

      for k, v in pairs db.indicators
        @frames1[unit]\add_indicator Elements.indicators[v]

    add_frame "player"
    add_frame "pet"
    add_frame "target"
    add_frame "targettarget"

  get_default_or_unit_setter: (fn) =>
    -- print "UnitFrames#get_default_or_unit_setter"

    (self, unit, ...) ->
      return fn unit, ... if unit ~= "default"

      for name, frame in pairs @frames
        fn name, ... if @db[name].defaults

  get_db_for_unit: (unit) =>
    -- print "UnitFrames#get_db_for_unit:#{unit}"

    db_unit = @db[unit]

    assert db_unit, "Unit not found: #{unit}"

    if db_unit.defaults
      @db.default
    else
      db_unit

  get_enabled_bars: (unit) =>
    db = @get_db_for_unit unit
    bars = {}
    table.insert bars, "health_bar" if db.bars.health_bar.enabled
    table.insert bars, "power_bar" if db.bars.power_bar.enabled
    table.insert bars, "cast_bar" if db.bars.cast_bar.enabled
    -- table.insert bars, "additional_power_bar" if db.bars.additional_power_bar.enabled
    bars

  get_weights_total: (unit) =>
    db = @get_db_for_unit unit
    idTweaksUtils.table.list.reduce ((a, v) -> a + db.bars[v].height), (@get_enabled_bars unit), 0

  size_position_frame: (unit, frame) =>
    -- print "UnitFrames#size_position_frame #{unit}"

    import layout from @db

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    switch layout
      when "horizontal"
        switch unit
          when "player"
            frame\ClearAllPoints!
            frame\SetPoint MR, UIParent, MC, -db.frame.size.margin / 2, -200
          when "target"
            frame\ClearAllPoints!
            frame\SetPoint ML, UIParent, MC, db.frame.size.margin / 2, -200
          when "targettarget"
            frame\ClearAllPoints!
            frame\SetPoint ML, @frames.target, MR, db.frame.size.margin, 0
          when "pet"
            frame\ClearAllPoints!
            frame\SetPoint MR, @frames.player, ML, -db.frame.size.margin, 0
          else
            print "Unknown unit: #{unit}"
      when "vertical"
        switch unit
          when "player"
            frame\ClearAllPoints!
            frame\SetPoint MC, UIParent, MC, 0, -200
          when "target"
            frame\ClearAllPoints!
            frame\SetPoint BC, @frames.player, TC, 0, db.frame.size.margin
          when "targettarget"
            frame\ClearAllPoints!
            frame\SetPoint BC, @frames.target, TC, 0, db.frame.size.margin
          when "pet"
            frame\ClearAllPoints!
            frame\SetPoint TC, @frames.player, BC, 0, -db.frame.size.margin
          else
            print "Unknown unit: #{unit}"
      else
        print "Unknown layout: #{layout}"

  size_position_elements: (unit, frame) =>
    -- print "UnitFrames#size_position_elements #{unit}"

    import layout from @db

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame\SetWidth db.frame.size.width
    frame\SetHeight db.frame.size.height

    switch layout
      when "horizontal"
        number_of_bars = #@get_enabled_bars unit
        weights_total = @get_weights_total unit

        width_minus_padding = db.frame.size.width - db.frame.size.padding * 2
        height_minus_padding = db.frame.size.height - db.frame.size.padding * number_of_bars + 1

        if db.bars.health_bar.enabled
          frame.Health\SetWidth width_minus_padding
          frame.Health\SetHeight height_minus_padding / weights_total * db.bars.health_bar.height
          frame.Health\ClearAllPoints!
          frame.Health\SetPoint TC, frame, TC, 0, -db.frame.size.padding

        if db.bars.power_bar.enabled
          frame.Power\SetWidth width_minus_padding
          frame.Power\SetHeight height_minus_padding / weights_total * db.bars.power_bar.height
          frame.Power\ClearAllPoints!
          frame.Power\SetPoint TC, frame.Health, BC, 0, -db.frame.size.padding

        if db.bars.cast_bar.enabled
          frame.Castbar\SetWidth width_minus_padding
          frame.Castbar\SetHeight height_minus_padding / weights_total * db.bars.cast_bar.height
          frame.Castbar.Icon\SetSize frame.Castbar\GetHeight!, frame.Castbar\GetHeight!
          frame.Castbar.Shield\SetSize frame.Castbar\GetHeight!, frame.Castbar\GetHeight!
          frame.Castbar\ClearAllPoints!
          frame.Castbar\SetPoint TC, frame.Power, BC, 0, -db.frame.size.padding

        if db.bars.additional_power_bar.enabled
          frame.AdditionalPower\SetWidth width_minus_padding
          frame.AdditionalPower\SetHeight height_minus_padding / weights_total * db.bars.additional_power_bar.height
          frame.AdditionalPower\ClearAllPoints!
          frame.AdditionalPower\SetPoint TC, frame.Power, BC, 0, -db.frame.size.padding

        -- TODO: Conditionally apply.
        frame.Buffs.size = db.frame.size.height / db.buffs.buffs.rows
        frame.Buffs.initialAnchor = BL
        frame.Buffs["growth-x"] = "RIGHT"
        frame.Buffs["growth-y"] = "UP"
        frame.Buffs\SetSize db.buffs.buffs.columns * frame.Buffs.size, db.buffs.buffs.rows * frame.Buffs.size
        frame.Buffs\ClearAllPoints!
        frame.Buffs\SetPoint BL, frame, TL, 0, db.frame.size.padding

        -- TODO: Conditionally apply.
        frame.Debuffs.size = db.frame.size.height / db.buffs.debuffs.rows
        frame.Debuffs.initialAnchor = TL
        frame.Debuffs["growth-x"] = "RIGHT"
        frame.Debuffs["growth-y"] = "DOWN"
        frame.Debuffs\SetSize db.buffs.debuffs.columns * frame.Debuffs.size, db.buffs.debuffs.rows * frame.Debuffs.size
        frame.Debuffs\ClearAllPoints!
        frame.Debuffs\SetPoint TL, frame, BL, 0, -db.frame.size.padding

      when "vertical"
        number_of_bars = #@get_enabled_bars unit
        weights_total = @get_weights_total unit

        width_minus_padding = db.frame.size.width - db.frame.size.padding * 2
        height_minus_padding = db.frame.size.height - db.frame.size.padding * number_of_bars + 1

        if db.bars.health_bar.enabled
          frame.Health\SetWidth width_minus_padding
          frame.Health\SetHeight height_minus_padding / weights_total * db.bars.health_bar.height
          frame.Health\ClearAllPoints!
          frame.Health\SetPoint TC, frame, TC, 0, -db.frame.size.padding

        if db.bars.power_bar.enabled
          frame.Power\SetWidth width_minus_padding
          frame.Power\SetHeight height_minus_padding / weights_total * db.bars.power_bar.height
          frame.Power\ClearAllPoints!
          frame.Power\SetPoint TC, frame.Health, BC, 0, -db.frame.size.padding

        if db.bars.cast_bar.enabled
          frame.Castbar\SetWidth width_minus_padding
          frame.Castbar\SetHeight height_minus_padding / weights_total * db.bars.cast_bar.height
          frame.Castbar.Icon\SetSize frame.Castbar\GetHeight!, frame.Castbar\GetHeight!
          frame.Castbar.Shield\SetSize frame.Castbar\GetHeight!, frame.Castbar\GetHeight!
          frame.Castbar\ClearAllPoints!
          frame.Castbar\SetPoint TC, frame.Power, BC, 0, -db.frame.size.padding

        if db.bars.additional_power_bar.enabled
          frame.AdditionalPower\SetWidth width_minus_padding
          frame.AdditionalPower\SetHeight height_minus_padding / weights_total * db.bars.additional_power_bar.height
          frame.AdditionalPower\ClearAllPoints!
          frame.AdditionalPower\SetPoint TC, frame.Power, BC, 0, -db.frame.size.padding

        -- TODO: Conditionally apply.
        frame.Buffs.size = db.frame.size.height / db.buffs.buffs.rows
        frame.Buffs.initialAnchor = TR
        frame.Buffs["growth-x"] = "LEFT"
        frame.Buffs["growth-y"] = "DOWN"
        frame.Buffs\SetSize db.buffs.buffs.columns * frame.Buffs.size, db.buffs.buffs.rows * frame.Buffs.size
        frame.Buffs\ClearAllPoints!
        frame.Buffs\SetPoint TR, frame, TL, -db.frame.size.padding, 0

        -- TODO: Conditionally apply.
        frame.Debuffs.size = db.frame.size.height / db.buffs.debuffs.rows
        frame.Debuffs.initialAnchor = TL
        frame.Debuffs["growth-x"] = "RIGHT"
        frame.Debuffs["growth-y"] = "DOWN"
        frame.Debuffs\SetSize db.buffs.debuffs.columns * frame.Debuffs.size, db.buffs.buffs.rows * frame.Debuffs.size
        frame.Debuffs\ClearAllPoints!
        frame.Debuffs\SetPoint TL, frame, TR, db.frame.size.padding, 0

      else
        print "Unknown layout: #{layout}"

  health_bar_set_texture: (unit, frame) =>
    -- print "UnitFrames#health_bar_set_texture:#{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.Health\SetStatusBarTexture db.bars.health_bar.texture

  health_bar_set_background: (unit, frame) =>
    -- print "UnitFrames#health_bar_set_background #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.Health.bg\SetTexture db.bars.health_bar.background
    frame.Health.bg.multiplier = 0.5

  power_bar_set_texture: (unit, frame) =>
    -- print "UnitFrames#power_bar_set_texture #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.Power\SetStatusBarTexture db.bars.power_bar.texture

  power_bar_set_background: (unit, frame) =>
    -- print "UnitFrames#power_bar_set_background #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.Power.bg\SetTexture db.bars.power_bar.background
    frame.Power.bg.multiplier = 0.5


  cast_bar_set_texture: (unit, frame) =>
    -- print "UnitFrames#cast_bar_set_texture #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.Castbar\SetStatusBarTexture db.bars.cast_bar.texture

  cast_bar_set_background: (unit, frame) =>
    -- print "UnitFrames#cast_bar_set_background #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.Castbar.bg\SetTexture db.bars.cast_bar.background
    frame.Castbar.bg.multiplier = 0.5

  additional_power_bar_set_texture: (unit, frame) =>
    -- print "UnitFrames#additional_power_bar_set_texture #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.AdditionalPower\SetStatusBarTexture db.bars.additional_power_bar.texture

  additional_power_bar_set_background: (unit, frame) =>
    -- print "UnitFrames#additional_power_bar_set_background #{unit}"

    frame = frame or @frames[unit]
    db = @get_db_for_unit unit

    frame.AdditionalPower.bg\SetTexture db.bars.additional_power_bar.background
    frame.AdditionalPower.bg.multiplier = 0.5

  create_background: (unit, frame) =>
    -- print "UnitFrames#create_background #{unit}"

    frame.background = frame\CreateTexture nil, "BACKGROUND"
    frame.background\SetAllPoints frame
    frame.background\SetColorTexture 0, 0, 0, 0.5

  create_healthbar: (unit, frame) =>
    -- print "UnitFrames#create_healthbar #{unit}"

    Health = CreateFrame "StatusBar", nil, frame
    Health.frequentUpdates = true
    Health.colorTapping = true
    Health.colorDisconnected = true
    Health.colorClass = true
    Health.colorReaction = true
    Health.colorHealth = true
    Health.bg = Health\CreateTexture nil, "BORDER"
    Health.bg\SetAllPoints Health

    frame.Health = Health

  create_powerbar: (unit, frame) =>
    -- print "UnitFrames#create_powerbar #{unit}"

    Power = CreateFrame "StatusBar", nil, frame
    Power.frequentUpdates = true
    Power.colorPower = true
    Power.bg = Power\CreateTexture nil, "BORDER"
    Power.bg\SetAllPoints Power

    frame.Power = Power

  create_cast_bar: (unit, frame) =>
    -- print "UnitFrames#create_cast_bar #{unit}"

    Castbar = CreateFrame "StatusBar", nil, frame
    Castbar.timeToHold = 0
    Castbar.hideTradeSkills = false

    Castbar.bg = Castbar\CreateTexture nil, "BACKGROUND"
    Castbar.bg\SetAllPoints Castbar

    Castbar.Spark = Castbar\CreateTexture nil, "OVERLAY"
    Castbar.Spark\SetBlendMode "ADD"
    Castbar.Spark\SetPoint MC, Castbar\GetStatusBarTexture!, MR, 0, 0

    Castbar.Time = Castbar\CreateFontString nil, "OVERLAY", "GameFontNormalSmall"
    Castbar.Time\SetPoint MR, Castbar, MR, 0, 0

    Castbar.Text = Castbar\CreateFontString nil, "OVERLAY", "GameFontNormalSmall"
    Castbar.Text\SetPoint ML, Castbar, ML, 0, 0

    Castbar.Icon = Castbar\CreateTexture nil, "OVERLAY"

    Castbar.Shield = Castbar\CreateTexture nil, "OVERLAY"
    Castbar.Shield\SetPoint MC, Castbar, MC, 0, 0

    Castbar.SafeZone = Castbar\CreateTexture nil, "OVERLAY"

    frame.Castbar = Castbar

  create_additional_powerbar: (unit, frame) =>
    -- print "UnitFrames#create_additional_powerbar #{unit}"

    AdditionalPower = CreateFrame "StatusBar", nil, frame
    AdditionalPower.frequentUpdates = true
    AdditionalPower.colorPower = true
    AdditionalPower.bg = AdditionalPower\CreateTexture nil, "BORDER"
    AdditionalPower.bg\SetAllPoints AdditionalPower

    frame.AdditionalPower = AdditionalPower

  create_buffs: (unit, frame) =>
    -- print "UnitFrames#create_buffs #{unit}"

    frame.Buffs = CreateFrame "Frame", nil, frame
    frame.Buffs.num = 16

  create_debuffs: (unit, frame) =>
    -- print "UnitFrames#create_debuffs #{unit}"

    frame.Debuffs = CreateFrame "Frame", nil, frame
    frame.Debuffs.num = 16

  create_nametext: (unit, frame) =>
    -- print "UnitFrames#create_nametext #{unit}"

    text_padding = 5
    frame.NameText = frame.Health\CreateFontString nil, "OVERLAY", "GameFontNormal"
    frame.NameText\SetFont frame.NameText\GetFont!, 9
    frame.NameText\SetTextColor 1, 1, 1, 1
    frame.NameText\SetPoint ML, frame.Health, ML, text_padding, 0
    frame.NameText\SetHeight 9
    frame\Tag frame.NameText, "[name][leader][status][threatcolor][threat]|r"

  create_healthtext: (unit, frame) =>
    -- print "UnitFrames#create_healthtext #{unit}"

    text_padding = 5
    frame.HealthText = frame.Health\CreateFontString nil, "OVERLAY", "GameFontNormal"
    frame.HealthText\SetFont frame.HealthText\GetFont!, 9
    frame.HealthText\SetTextColor 1, 1, 1, 1
    frame.HealthText\SetPoint MR, frame.Health, MR, -text_padding, 0
    frame.HealthText\SetHeight 9
    frame\Tag frame.HealthText, "[curhp]"

  create_classpowertext: (unit, frame) =>
    -- print "UnitFrames#create_classpowertext #{unit}"

    text_padding = 5
    frame.ClassPowerText = frame.Power\CreateFontString nil, "OVERLAY", "GameFontNormal"
    frame.ClassPowerText\SetFont frame.ClassPowerText\GetFont!, 9
    frame.ClassPowerText\SetTextColor 1, 1, 1, 1
    frame.ClassPowerText\SetPoint ML, frame.Power, ML, text_padding, 0
    frame.ClassPowerText\SetHeight 9
    frame\Tag frame.ClassPowerText, "[runes][arcanecharges][chi][holypower][soulshards][cpoints]"

  create_powertext: (unit, frame) =>
    -- print "UnitFrames#create_powertext #{unit}"

    text_padding = 5
    frame.PowerText = frame.Power\CreateFontString nil, "OVERLAY", "GameFontNormal"
    frame.PowerText\SetFont frame.PowerText\GetFont!, 9
    frame.PowerText\SetTextColor 1, 1, 1, 1
    frame.PowerText\SetPoint MR, frame.Power, MR, -text_padding, 0
    frame.PowerText\SetHeight 9
    frame\Tag frame.PowerText, "[curpp]"

  create_unit: (unit, frame) =>
    @create_background unit, frame
    @create_healthbar unit, frame
    @create_powerbar unit, frame
    @create_cast_bar unit, frame
    @create_additional_powerbar unit, frame
    @create_buffs unit, frame
    @create_debuffs unit, frame
    @create_nametext unit, frame
    @create_healthtext unit, frame
    @create_classpowertext unit, frame
    @create_powertext unit, frame

  update_unit: (unit, frame) =>
    db = @get_db_for_unit unit
    @size_position_elements unit, frame
    @size_position_frame unit, frame

    if db.bars.health_bar.enabled
      @health_bar_set_texture unit, frame
      @health_bar_set_background unit, frame

    if db.bars.power_bar.enabled
      @power_bar_set_texture unit, frame
      @power_bar_set_background unit, frame

    if db.bars.cast_bar.enabled
      @cast_bar_set_texture unit, frame
      @cast_bar_set_background unit, frame

    if db.bars.additional_power_bar.enabled
      @additional_power_bar_set_texture unit, frame
      @additional_power_bar_set_background unit, frame

  update_all_units: =>
    for i, unit in ipairs @units
      @update_unit unit, @frames[unit]

  layout_default: (unit, frame) =>
    -- print "UnitFrames#layout_default #{unit}"

    @create_unit unit, frame
    @update_unit unit, frame

idTweaks\add_module UnitFrames idTweaks
